#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <time.h>

#include "pngfuncs.h"

#include "r300_reg.h"
#include "r300_emit.h"

//#define ADDR 0xe9000000
//#define SIZE (64*1024)
#define AGP_ADDR 0xd0000000
#define AGP_SIZE (128*1024*1024)
//#define FB_ADDR 0xd8000000
//#define FB_SIZE (128*1024*1024)
#define RB_OFFSET 0x00001000

extern off_t addr;
extern size_t size;
off_t agp_addr = AGP_ADDR;
size_t agp_size = AGP_SIZE;
extern off_t fb_addr;
extern size_t fb_size;
int rb_offset = RB_OFFSET;

// File pointers to the memory
int mem, agp_mem, fb_mem;
// Handles to the shared memory of the graphics card
unsigned long *prt, *ring_prt, *agp_prt, *fb_prt;

void nrt_analyzer(unsigned long *addr, unsigned long size, int ib);
void rt_analyzer(unsigned long *addr, unsigned long size, int ib);

/**
 * This method dumps indirect buffers to according files. The files will be
 * named like ib_(INDERECT_BUFFER)_(SIZE). The part (INDERECT_BUFFER) and (SIZE)
 * come from the parameters.
 *
 * \param addr Address of the indirect buffer
 * \param size Size of the indirect buffer (used as SIZE in the filename)
 * \param ib The indirect buffer (used in the filename as INDERECT_BUFFER)
 */
void dump_indirect(unsigned long *addr, unsigned long size, int ib) {
	int i;
	FILE *fout;
	char filename[1024];

	sprintf(filename, "ib_%04d_%04lu", ib, size);

	fout = fopen(filename, "w");

	for(i = 0; i < size; addr++, i++){
		fprintf(fout, "%08x\n", *addr);
	}

	fclose(fout);
}

#define IB_ARCHIVE_SIZE (1024*1024*128)
#define IB_ARCHIVE_MAX (5000) /*(1024)*/

#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

void parse_vfp(unsigned long *addr, unsigned long len, int ib);

int main(int argc, char *argv[]) {
	unsigned long ring_head, ring_tail, ring_size;
	unsigned long index, ib_count=0;
	unsigned long ib_addr, ib_size;
	FILE *fout, *gc;
	unsigned long *ib_archive;
	unsigned long ib_eidx[IB_ARCHIVE_MAX]={ 0 };
	time_t t_last_ib, t_started;
	//unsigned char *foo;
	unsigned long *mapped_ib_addr;
	unsigned long proc, proc_temp;

	unsigned long *temp;
	unsigned long temp_size;
	char text[100];
	int i;

	// Check if we were called with 2 arguments
	if (argc != 3) {
		fprintf(stderr, "No arguments given!\n");
		fprintf(stdout, "Usage\n");
		fprintf(stdout, "=====\n");
		fprintf(stdout, "./iba --log DIRECTORY\n");
		fprintf(stdout, "DIRECTORY is the name of a directory\n");
		fprintf(stdout, "(it will contain the logfiles once we are done)\n\n");
		fprintf(stdout, "./iba --analyze LOGFILE\n");
		fprintf(stdout, "LOGFILE is a file captured with ./iba --log\n");
		return 2;
	}

	// Analyze mode
	if (strcmp(argv[1],"--analyze") == 0) {

		// Open the given file
		fout = fopen(argv[2], "r");
		if (fout == NULL) {
			fprintf(stderr, "Could not open %s for reading.\n", argv[1]);
			return 3;
		}

		temp=malloc(1024*sizeof(unsigned long));
		memset(temp, 0, 1024*sizeof(unsigned long));
		temp_size = 0;

		// Read file fout into variable text
		while(fgets(text, 100, fout) != NULL)
			sscanf(text, "%x\n", &(temp)[temp_size++]);

		fprintf(stderr, "%d dwords\n", temp_size);
		for(i=0; i < temp_size; i++)
			fprintf(stderr, "%08x\n", temp[i]);

		parse_vfp(temp, temp_size, 0);

		return 0;
	} else if (strcmp(argv[1],"--log") != 0) {
		fprintf(stderr, "Unkown command given. Exiting\n");
	}

	// Log Mode
	if(mkdir(argv[2], 0777) == -1)
		perror("mkdir:");
	if(chdir(argv[2]) == -1) {
		perror("chdir:");
		return 4;
	}

	/*
	foo=malloc(1024*1024*3);
	for(index=0; index < 1024*1024*3; index+=3){
		foo[index]=index;
		foo[index+1]=128;
		foo[index+2]=128;
	}
	write_png_file("test.png", foo, 1024, 1024, 3);
	return 0;
	*/

	ib_archive = malloc(IB_ARCHIVE_SIZE*sizeof(unsigned long));

	if (get_pci() != 0) {
		printf("Could not find a card to use ... exiting.\n");
		return 0;
	}

	printf("Setting up mem %0x, %d B\n", addr, size);
	// Setup variable prt*
	if((mem = open("/dev/mem", O_RDWR/*O_RDONLY*/)) == -1){
		printf("Could not open /dev/mem. You're not running this as root?\n");
		return -1;
	}
	prt = mmap(NULL, size, PROT_READ|PROT_WRITE, MAP_SHARED, mem, addr);
	if (prt == NULL) {
		printf("Could not mmap /dev/mem for ptr.\n");
		goto close_ptr;
	}

	printf("Setting up agp_mem %0x, %d B\n", agp_addr, agp_size);
	// Setup variable agp_prt*
	if((agp_mem = open("/dev/mem" /*"/dev/agpgart"*/, O_RDWR)) == -1){
		printf("Could not open /dev/mem for AGP access.\n");
		goto unmap_ptr;
	}
	agp_prt = mmap(NULL, agp_size, PROT_READ | PROT_WRITE, MAP_SHARED, agp_mem, agp_addr);
	if (agp_prt == NULL) {
		printf("Could not mmap /dev/mem for agp_prt.\n");
		goto close_agp_ptr;
	}

	printf("Setting up fb_mem %0x, %d B\n", fb_addr, fb_size);
	// Setup variable fb_prt
	if((fb_mem = open("/dev/mem" /*"/dev/agpgart"*/, O_RDWR)) == -1){
		printf("Could not open /dev/mem for Framebuffer access.\n");
		goto unmap_agp_ptr;
	}
	fb_prt = mmap(NULL, fb_size, PROT_READ | PROT_WRITE, MAP_SHARED, fb_mem, fb_addr);
	if (fb_prt == NULL) {
		printf("Could not mmap /dev/mem for fb_prt.\n");
		goto close_fb_ptr;
	}


#if 0
	// Testwrite 0xdeadbeef to the AGP
	agp_prt[(68*1024*1024) >> 2] = 0xdeadbeef;
	if(agp_prt[(68*1024*1024) >> 2] != 0xdeadbeef)
		fprintf(stderr, "APG write test failed!\n");
	else
		fprintf(stderr, "APG write test ok\n");
	
	return 0;
#endif

	ring_prt = ((unsigned long *)((char *)agp_prt)+rb_offset);

	fout = fopen("dump", "w");
	gc = fopen("gc", "w");

	printf("addr=0x%08x\n", prt[RADEON_CP_RB_RPTR_ADDR>>2]);
	ring_head=ring_tail=prt[RADEON_CP_RB_RPTR>>2];

	ring_size=(1 << ((prt[0x0704>>2] & 0x000000ff)+1));

	printf("ring size 0x%08x dwords\n", ring_size);

	t_last_ib=t_started=time(NULL);
	for(ib_count=0;;){
		if(t_last_ib < time(NULL)-2 && t_last_ib != t_started){
			printf("timed out\n");
			break;
		}
		ring_tail=prt[RADEON_CP_RB_RPTR>>2];

		if(ring_tail == ring_head)
			continue;

		//if((ring_tail-ring_head)%16) printf("invalid size\n");

		//fprintf(stderr, "reading...\n");
		for(index=ring_head; index < ring_tail; index+=proc/*++*/, index &= ring_size-1){
			//fflush(fout);
			/*
			if(ring_prt[index] == 0x80000000){
				continue;
			}
			*/
#if 1
			if(ring_prt[index] == CP_PACKET0(CP_IB_BASE,1)){
				//fprintf(stderr, "ib!\n");
				/*index++;*/ ib_addr=ring_prt[index+1];
				/*index++;*/ ib_size=ring_prt[index+2];
				
				mapped_ib_addr=((unsigned long *)(char *)(agp_prt) + (ib_addr - agp_addr));
						
				memcpy(ib_archive+ib_eidx[ib_count], mapped_ib_addr, ib_size*sizeof(unsigned long));
				rt_analyzer(mapped_ib_addr, ib_size, ib_count);
				ib_count++;
				ib_eidx[ib_count] = ib_eidx[ib_count-1] + ib_size;
				
				
				//printf("indirect addr=0x%08x size=%lu\n", ib_addr, ib_size);
				t_last_ib=time(NULL);
				
				if(ib_count >= IB_ARCHIVE_MAX-1)
					goto exit;
	
				//continue;
				proc=3;
			}else 
#endif
				proc=1;
			if(/*t_last_ib != t_started*/1){
				for(proc_temp=0; proc_temp < proc; proc_temp++)
					fprintf(fout, "%08x\n", ring_prt[ (index+proc_temp) & (ring_size-1) ]);
				fflush(fout);
				//t_last_ib=time(NULL);
				//prev_packet2=0;
			}
		}		

		ring_head=ring_tail;
	}
	exit:
	nrt_analyzer(ib_archive, ib_eidx[ib_count], 0);
	printf("saving buffers\n");
	for(index=0; index < ib_count-1; index++){
		mapped_ib_addr=ib_archive+ib_eidx[index];
		ib_size=ib_eidx[index+1]-ib_eidx[index];
		
		//nrt_analyzer(mapped_ib_addr, ib_size, index);
		dump_indirect(mapped_ib_addr, ib_size, index);
	}
	fclose(gc);
	fclose(fout);

	// Close fb_prt
	if(munmap(fb_prt, fb_size) == -1)
		perror("munmap");
close_fb_ptr:
	close(fb_mem);

	// Close agp_ptr
unmap_agp_ptr:
	if(munmap(agp_prt, agp_size) == -1)
		perror("munmap");
close_agp_ptr:
	close(agp_mem);

	// Close ptr
unmap_ptr:
	if(munmap(prt, size) == -1)
		perror("munmap");
close_ptr:
	close(mem);

	return 0;
}
