#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "pngfuncs.h"

// R300 Includes
#include "r300_reg.h"
#include "r300_emit.h"

#include "vertex_shader.h"

#define UNK1 0x80
#define UNK1_SHIFT (24)

#define R300_TX_UNK_0                      0x4420
#define R300_TX_UNK_1                      0x4460
#define R300_TX_UNK_2                      0x44A0
#define R300_TX_UNK_3                      0x44E0

#define MKN0(name) { R300_VPI_OUT_OP_##name, #name}
struct{
	int id;
	char *name;
}VPI_OUT_OP[]={
MKN0(DOT),
MKN0(MUL),
MKN0(ADD),
MKN0(MAD),
MKN0(FRC),
MKN0(MAX),
MKN0(MIN),
MKN0(SGE),
MKN0(SLT),
MKN0(EXP),
MKN0(LOG),
MKN0(LIT),
MKN0(POW),
MKN0(RCP),
MKN0(RSQ),
MKN0(EX2),
MKN0(LG2),
MKN0(MAD_2),
};

#define MKN1(name) { VSF_OUT_CLASS_##name, #name}
struct{
	int id;
	char *name;
}VSF_OUT_CLASS[]={
MKN1(TMP),
MKN1(RESULT),
};

#define MKN2(name) { VSF_FLAG_##name, #name}
struct{
	int id;
	char *name;
}VSF_FLAG[]={
MKN2(X),
MKN2(Y),
MKN2(Z),
MKN2(W),
MKN2(NONE),
};


#define MKN3(name) { VSF_IN_CLASS_##name, #name}
struct{
	int id;
	char *name;
}VSF_IN_CLASS[]={
MKN3(TMP),
MKN3(ATTR),
MKN3(PARAM),
MKN3(NONE),
};
int indirect_end, indirect_size;

#define MKN4(name) { VSF_IN_COMPONENT_##name, #name}
struct{
	int id;
	char *name;
}VSF_IN_COMPONENT[]={
MKN4(X),
MKN4(Y),
MKN4(Z),
MKN4(W),
MKN4(ZERO),
MKN4(ONE),
};

#define MKN5(name) { R300_VPI_IN_NEG_##name, #name}
struct{
	int id;
	char *name;
}VPI_IN_NEG[]={
MKN5(X),
MKN5(Y),
MKN5(Z),
MKN5(W),
};


FILE *nout=NULL;
/*
unsigned long get_align(unsigned long addr){
	unsigned long align=0;
	int index;
			
	for(index=0; index < 32; index++){
		if(addr	& (1<<index))
			break;
		align |= (1<<index);
	}
	return align;
}
*/
void parse_op(unsigned long d) {
	int temp;
	int op_found=0;
	if (nout == NULL) {
		nout = fopen("nrta", "w");
		if (nout == NULL) {
			perror("nrt_analyzer:");
			exit(-1);
		}
	}
	for(temp=0; temp < sizeof(VPI_OUT_OP)/sizeof(*VPI_OUT_OP); temp++)
		if(VPI_OUT_OP[temp].id == (d & 0xff) ){
			fprintf(nout, "%s ", VPI_OUT_OP[temp].name);
			op_found=1;
			break;
		}
		
	if(!op_found){
		fprintf(nout, "unk(%d)", (d & 0xff));
		//exit(-1);
	}
	
	for(temp=0; temp < sizeof(VSF_OUT_CLASS)/sizeof(*VSF_OUT_CLASS); temp++)
		if(VSF_OUT_CLASS[temp].id == ((d >> 8) & 0x2) )
			fprintf(nout, "%s ", VSF_OUT_CLASS[temp].name);
	
	fprintf(nout, "%d.", /*(d & 31)*/ (d >> 13) & 31);
		
	for(temp=0; temp < sizeof(VSF_FLAG)/sizeof(*VSF_FLAG); temp++)
		if(VSF_FLAG[temp].id & ((d>>20) & 0xf) )
			fprintf(nout, "%s ", VSF_FLAG[temp].name);
}
		
void parse_source(unsigned long d){
int temp;
int class_found=0;
	
	for(temp=0; temp < sizeof(VSF_IN_CLASS)/sizeof(*VSF_IN_CLASS); temp++)
		if(VSF_IN_CLASS[temp].id == (d & 0xf) ){
			fprintf(nout, "%s ", VSF_IN_CLASS[temp].name);
			class_found=1;
			break;
		}
	if(!class_found)
		return;
	
	fprintf(nout, "%d", (d & R300_VPI_IN_REG_INDEX_MASK) >> R300_VPI_IN_REG_INDEX_SHIFT );
	
	/*fprintf(nout, "{");
	for(temp=0; temp < sizeof(VSF_FLAG)/sizeof(*VSF_FLAG); temp++)
		if(VSF_FLAG[temp].id & ((d >> 25) & 0xf) )
			fprintf(nout, "%s ", VSF_FLAG[temp].name);
	fprintf(nout, "} ");*/
		
	fprintf(nout, "{");
	for(temp=0; temp < sizeof(VSF_IN_COMPONENT)/sizeof(*VSF_IN_COMPONENT); temp++)
		if(VSF_IN_COMPONENT[temp].id == ((d >> R300_VPI_IN_X_SHIFT) & 0x7) )
			fprintf(nout, "%s ", VSF_IN_COMPONENT[temp].name);

	for(temp=0; temp < sizeof(VSF_IN_COMPONENT)/sizeof(*VSF_IN_COMPONENT); temp++)
		if(VSF_IN_COMPONENT[temp].id == ((d >> R300_VPI_IN_Y_SHIFT) & 0x7) )
			fprintf(nout, "%s ", VSF_IN_COMPONENT[temp].name);
		
	for(temp=0; temp < sizeof(VSF_IN_COMPONENT)/sizeof(*VSF_IN_COMPONENT); temp++)
		if(VSF_IN_COMPONENT[temp].id == ((d >> R300_VPI_IN_Z_SHIFT) & 0x7) )
			fprintf(nout, "%s ", VSF_IN_COMPONENT[temp].name);
		
	for(temp=0; temp < sizeof(VSF_IN_COMPONENT)/sizeof(*VSF_IN_COMPONENT); temp++)
		if(VSF_IN_COMPONENT[temp].id == ((d >> R300_VPI_IN_W_SHIFT) & 0x7) )
			fprintf(nout, "%s", VSF_IN_COMPONENT[temp].name);
	
	fprintf(nout, "} ");
	
	for(temp=0; temp < sizeof(VPI_IN_NEG)/sizeof(*VPI_IN_NEG); temp++)
		if(VPI_IN_NEG[temp].id & d)
			fprintf(nout, "neg %s ", VPI_IN_NEG[temp].name);

}

void parse_vfp(unsigned long *addr, unsigned long len, int ib){
	int i;
	//int temp;
	fprintf(nout, "\nprogram:\n");

	for (i = 0; i < len;){
		parse_op(addr[i]); i++;
		parse_source(addr[i]); i++;
		parse_source(addr[i]); i++;
		parse_source(addr[i]); i++;
		fprintf(nout, "\n");
	}
}

float to_float(unsigned long p){
	float *temp = (float*) &p;
	return *temp;
}

void print_one(unsigned long p){
	float *f;
	char text[100]={ 0 };
	int i;
	
	fprintf(nout, "%08x", p);
	f=(float *)(&p);
	if(p & 0xf0000000)
		sprintf(text, "(%f)", *f);
	
	for(i=0; i < 11; i++)
		if(text[i] == '\0')
			text[i]=' ';
	
	fprintf(nout, "%s", text);
}

/**
 * Non-Realtime Analyzer
 *
 * \param addr Address of the indirect buffer
 * \param size Size of the indirect buffer (used as SIZE in the filename)
 * \param ib The indirect buffer (used in the filename as INDERECT_BUFFER)
 */
void nrt_analyzer(unsigned long *addr, unsigned long size, int ib) {
	int i;
	//unsigned long *p;
	//int tmu;
	//unsigned long t_width=0, t_height=0;

	static long pvs_addr=-1;
	unsigned long pvs_len;
	//unsigned long left;
	//float f;
	//unsigned long pfs_params;
	//unsigned long packet_len;
	//int index;

#define POS_PRESENT (2<<20)
#define COLOR_PRESENT (1<<20)
#define BOTH_COLOR_PRESENT (4<<20)
#define TEX_COORD_PRESENT (4<<20)

	if(nout == NULL) {
		nout = fopen("nrta", "w");
		if (nout == NULL) {
			perror("nrt_analyzer:");
			exit(-1);
		}
	}

	//parse_op(0x3f7ae148);
	//return ;

	for(i=0; i < size; i++){
#if 0
		if(addr[i] == 0x000110f8){ /* line stipple */
			i++; fprintf(nout, "%d ", addr[i]);
			i++; fprintf(nout, "%d\n", addr[i]);
		}
#endif

#if 0
		if(addr[i] == RADEON_CP_PACKET2)
			continue;
		if(addr[i] == 0x0)
			continue;
		
		packet_len=(addr[i]>>16)+1;
		
		fprintf(nout, "packet len %d\n", packet_len);
		if(packet_len > 1){
			for(index=0; index < packet_len; index++)
				fprintf(nout, "\n");
				
		}
		for(; packet_len > 0 ; i++, packet_len--){
			if(!(i < size)){
				printf("broken packet\n");
				break;
			}
				
		}
		/*
		if((addr[i] & CP_PACKET0(R300_PFS_PARAM_0_X, 0)) == CP_PACKET0(R300_PFS_PARAM_0_X, 0)){
			pfs_params=addr[i]>>16;
			
			fprintf(nout, "pfs_params %d:\n", pfs_params);
			for(; i < size && pfs_params > 0; i++, pfs_params--){
				fprintf(nout, "%08x ", addr[i]);
				f=((float *)addr)[i];
				if(addr[i] & 0xf0000000)
					fprintf(nout, "%f", f);
				fprintf(nout, "\n");
			}
			printf("\n");
		}
		*/
		/*
		if(addr[i] == 0x0007 1300){
			R300_PFS_PARAM_0_X
		}
		*/
#endif
		if(addr[i] == 0x000310a9){
			i++; fprintf(nout, "%f(%f) ", to_float(addr[i]), to_float(addr[i])/12.0);
			i++; fprintf(nout, "%f(%f) ", to_float(addr[i]), to_float(addr[i])/4.0);
			i++; fprintf(nout, "%f(%f) ", to_float(addr[i]), to_float(addr[i])/12.0);
			i++; fprintf(nout, "%f(%f)\n", to_float(addr[i]), to_float(addr[i])/4.0);
		}
#if 1
		if(addr[i] == 0x000208b4 /*CP_PACKET0(R300_VAP_PVS_CNTL_1, 2)*/ ){
			i++; //fprintf(nout, "R300_VAP_PVS_CNTL_1=%08x\n", addr[i]);
			fprintf(nout, "program start %d\n", (addr[i] >> R300_PVS_CNTL_1_PROGRAM_START_SHIFT) & 0x3ff);
			fprintf(nout, "program unknown_0 %d\n", (addr[i] >> R300_PVS_CNTL_1_POS_END_SHIFT) & 0x3ff);
			fprintf(nout, "program len %d\n", (addr[i] >> R300_PVS_CNTL_1_PROGRAM_END_SHIFT) & 0x3ff);
			
			i++; //fprintf(nout, "R300_VAP_PVS_CNTL_2=%08x\n", addr[i]);
			fprintf(nout, "program param start %d\n", (addr[i] >> R300_PVS_CNTL_2_PARAM_OFFSET_SHIFT) & 0xffff);
			fprintf(nout, "program param count %d\n", (addr[i] >> R300_PVS_CNTL_2_PARAM_COUNT_SHIFT) & 0xffff);
			
			i++; //fprintf(nout, "R300_VAP_PVS_CNTL_3=%08x\n", addr[i]);
			fprintf(nout, "program unknown2 %d\n", (addr[i] >> R300_PVS_CNTL_3_PROGRAM_UNKNOWN2_SHIFT) & 0x3ff);
			fprintf(nout, "program unknown %d\n", (addr[i] >> R300_PVS_CNTL_3_PROGRAM_UNKNOWN_SHIFT) & 0x3ff);
			
			continue;
		}
		
		if(addr[i] == CP_PACKET0(R300_VAP_PVS_UPLOAD_ADDRESS, 0) ){
			i++; pvs_addr=addr[i];
			continue;
		}

		if( (addr[i] & CP_PACKET0(R300_VAP_PVS_UPLOAD_DATA, 0)) == CP_PACKET0(R300_VAP_PVS_UPLOAD_DATA, 0) && pvs_addr != -1){
			pvs_len=(addr[i] >> 16)+1;
			i++;
			
			if(pvs_addr /*== 0*/< 0x200){
				fprintf(nout, "at addr %x", pvs_addr);
				parse_vfp(addr+i, pvs_len, ib);
				addr+=pvs_len;
			}else{
				int value_index=0;
				
				fprintf(nout, "pvs_addr %x len %d:\n", pvs_addr, pvs_len);
				for(; i < size && pvs_len > 0; i++, pvs_len--, value_index++){
					if((pvs_len % 4)==0){
						fprintf(nout, "\n");
						fprintf(nout, "param %02d:", value_index/4);
					}
					print_one(addr[i]);
						
				}
				fprintf(nout, "\n");
			}
			i--;
			pvs_addr=-1;
			continue;
		}
		
#endif		
		/*if(addr[i]==0x80008000)
			fprintf(gc, "index %lu ib %d\n", i, curr_ib);*/
		
#if 0
		for(tmu=0; tmu < 16; tmu++){
			
			if(addr[i] == CP_PACKET0(R300_TX_UNK_0+tmu*4,0)){
				i++; fprintf(nout, "R300_TX_UNK_0=%08x ib %d tmu %d\n", addr[i], ib, tmu); continue;
			}	
			if(addr[i] == CP_PACKET0(R300_TX_UNK_1+tmu*4,0)){
				i++; fprintf(nout, "R300_TX_UNK_1=%08x ib %d tmu %d\n", addr[i], ib, tmu); continue;
			}	
			if(addr[i] == CP_PACKET0(R300_TX_UNK_2+tmu*4,0)){
				i++; fprintf(nout, "R300_TX_UNK_2=%08x ib %d tmu %d\n", addr[i], ib, tmu); continue;
			}	
			if(addr[i] == CP_PACKET0(R300_TX_UNK_3+tmu*4,0)){
				i++; fprintf(nout, "R300_TX_UNK_3=%08x ib %d tmu %d\n", addr[i], ib, tmu); continue;
			}	
			
			if(addr[i] == CP_PACKET0(R300_TX_FILTER_0+tmu*4,0)){
				i++; fprintf(nout, "R300_TX_FILTER_0=%08x ib %d tmu %d\n", addr[i], ib, tmu); continue;
			}
			
			if(addr[i] == CP_PACKET0(0x4500+tmu*4,0)){
				i++; fprintf(nout, "R300_TX_FILTER_0=%08x ib %d tmu %d\n", addr[i], ib, tmu); continue;
			}
			if(addr[i] == CP_PACKET0(R300_TX_FILTER1_0+tmu*4,0)){
				i++; fprintf(nout, "R300_TX_FILTER1_0=%08x ib %d tmu %d\n", addr[i], ib, tmu); continue;
			}	
			if(addr[i] == CP_PACKET0(R300_TX_SIZE_0+tmu*4,0)){
				i++;
				t_width=addr[i] & 0x7ff;
				t_height=(addr[i]>>11) & 0x7ff;
				fprintf(nout, "tex width %lu, height %lu, ib %d tmu %d\n", t_width+1, t_height+1, ib, tmu);
				//fprintf(nout, "R300_TX_SIZE_0=%08x ib %d tmu %d\n", addr[i], ib, tmu);
				continue;
			}
			if(addr[i] == CP_PACKET0(R300_TX_OFFSET_0+tmu*4,0)){
				i++; fprintf(nout, "R300_TX_OFFSET_0=%08x ib %d tmu %d\n", addr[i], ib, tmu);
			}
			if(addr[i] == CP_PACKET0(R300_TX_FORMAT_0+tmu*4,0)){
				i++; fprintf(nout, "R300_TX_FORMAT_0=%08x ib %d tmu %d\n", addr[i], ib, tmu); continue;
			}	
			if(addr[i] == CP_PACKET0(R300_TX_CHROMA_KEY_0+tmu*4,0)){
				i++; fprintf(nout, "R300_TX_CHROMA_KEY_0=%08x ib %d tmu %d\n", addr[i], ib, tmu); continue;
			}	
			if(addr[i] == CP_PACKET0(R300_TX_BORDER_COLOR_0+tmu*4,0)){
				i++; fprintf(nout, "R300_TX_BORDER_COLOR_0=%08x ib %d tmu %d\n", addr[i], ib, tmu); continue;
			}
		}
#endif
	}
}
