#include <stdio.h>
#include <stdlib.h>

#include "pngfuncs.h"

#include "r300_reg.h"
#include "r300_emit.h"


#define TEX_BYTES 4

static FILE *fout=NULL;

// Handles to the shared memory of the graphics card
extern unsigned long *prt, *ring_prt, *agp_prt, *fb_prt;

void check_texture(unsigned long addr, unsigned long width, unsigned long height) {
	unsigned char *texture_addr;
	char text[100];
	//unsigned char *buffer;
	//unsigned long curr_width, curr_height;
	//int index;
	//int total_size;
	//unsigned char curr_value;
	//unsigned long prev_tex_index, tex_index;

	static int tex_i=0;

	//if(tex_i > 5)
		return ;

	fprintf(stderr, "at %08x, width=%d, height=%d\n", addr, width, height);

	texture_addr = (unsigned char*)fb_prt;
	texture_addr += addr & 0x0fffffff;
	srand(0);
	//sleep(1);
	//width=32;
	
#if 0
	curr_width=width;
	curr_height=height;
	total_size=0;
	printf("size %lu x %lu\n", width, height);
	
	while(curr_width > 0 && curr_height > 0){
		total_size += curr_width * curr_height * TEX_BYTES;	
#if 0
		curr_width >>= 1;
		curr_height >>= 1;
#else
		curr_width = curr_height = 0;
#endif
	}
#endif
	sprintf(text, "texture_%d.png", tex_i);
	write_png_file(text, texture_addr, width, height*4, 4);
	
	/*sprintf(text, "texture_mipmap_%d.png", 0);
	write_png_file(text, texture_addr, width/2, height  *4, 4);
	exit(0);*/
	tex_i++;
}

/**
 * This method checks the type of a primitive and returns a human readable
 * name.
 * \param type The type of the prim (i.e. R300_VAP_VF_CNTL__PRIM_TRIANGLES)
 * \return Returns a human reable name for that type (i.e. "triangles")
 */
char *prim_name(unsigned char type) {
	switch (type) {
		case R300_VAP_VF_CNTL__PRIM_NONE: return "none";
		case R300_VAP_VF_CNTL__PRIM_POINTS: return "points";
		case R300_VAP_VF_CNTL__PRIM_LINES: return "lines";
		case R300_VAP_VF_CNTL__PRIM_LINE_STRIP: return "line strip";
		case R300_VAP_VF_CNTL__PRIM_TRIANGLES: return "triangles";
		case R300_VAP_VF_CNTL__PRIM_TRIANGLE_FAN: return "triangle fan";
		case R300_VAP_VF_CNTL__PRIM_TRIANGLE_STRIP: return "triangle strip";
		case R300_VAP_VF_CNTL__PRIM_LINE_LOOP: return "line loop";
		case R300_VAP_VF_CNTL__PRIM_QUADS: return "quads";
		case R300_VAP_VF_CNTL__PRIM_QUAD_STRIP: return "quad strip";
		case R300_VAP_VF_CNTL__PRIM_POLYGON: return "polygon";
	}
	return "unknown";
}

void dump_idx_buf_c(unsigned long addr, unsigned long size){
	int i;
	unsigned char *buf;
	
	buf=((char *)agp_prt)+(addr-0xd0000000);
	
	fprintf(fout, "<8bit ");
	for(i=0; i < size*2; i++)
		fprintf(fout, "%d ", buf[i]);
	
	fprintf(fout, ">");
	/*
	for(; i < size+10; i++)
		fprintf(fout, "%d ", buf[i]);
	*/
	//fprintf(fout, "\n");
}

void dump_idx_buf(unsigned long addr, unsigned long size){
	int i;
	unsigned short int *buf;
	
	buf=((unsigned short int *)((char *)agp_prt)+(addr-0xd0000000));
	
	fprintf(fout, "<16bit ");
	for(i=0; i < size; i++)
		fprintf(fout, "%d ", buf[i]);
	
	fprintf(fout, ">");
	
	for(; i < size+10; i++)
		fprintf(fout, "%d ", buf[i]);
	
	//fprintf(fout, "\n");
}

void dump_idx_buf_l(unsigned long addr, unsigned long size){
	int i;
	unsigned long *buf;
	//return ;
	buf=((unsigned long *)((char *)agp_prt)+(addr-0xd0000000));
	
	fprintf(fout, "<32bit ");
	for(i=0; i < size/2; i++){
		//parse_op(buf[i]);
		fprintf(fout, "%x ", buf[i]);
		//fprintf(fout, "\n");
	}	
	fprintf(fout, ">");
	/*
	for(; i < size+10; i++)
		fprintf(fout, "%d ", buf[i]);
	*/
	//fprintf(fout, "\n");
}

void check_magic_1(unsigned long addr, unsigned long elements){
	unsigned char magic1_tbl[4]={ 0, 6, 4, 2 };
	unsigned long magic_tbl2[6]= { 0x3f51eb85, 0x3e99999a, 0x3f1eb852, 0x3e6b851f, 0x3f2b851f, 0x3eb33333 };
	unsigned long magic_tbl3[1]= { 0xbf000000 };
	unsigned long magic_1;
	int i;
	unsigned long *buf;

	buf=((unsigned long *)((char *)agp_prt)+(addr-0xd0000000));

	magic_1=magic1_tbl[elements % 4];
			
	if(elements == 1){
		if(buf[0] != magic_tbl3[0]) goto fail;
		if(buf[1] != 0) goto fail;
		if(buf[2] != 0) goto fail;
		
		if(buf[3] != magic_tbl2[3]) goto fail;
		if(buf[4] != magic_tbl2[4]) goto fail;
		if(buf[5] != magic_tbl2[5]) goto fail;
	}else{
		for(i=0; i < magic_1; i++)
			if(buf[i] != magic_tbl2[6-magic_1+i])
				goto fail;
	}
	fprintf(fout, "magic_1 ok ");
	return ;
	
	fail:
	fprintf(fout, "*magic_1 fails* ");
}

unsigned long get_align(unsigned long addr){
	unsigned long align=0;
	int index;
			
	for(index=0; index < 32; index++){
		if(addr	& (1<<index))
			break;
		align |= (1<<index);
	}
	return align;
}

typedef unsigned long long /*int*/ dtype;
#define DTYPE_LAST (1<<63)
#define DTYPE_BITS 64

/**
 * Prints the given num to fout
 */
void print_bits(dtype num) {
	int index;
	int sp=0;
	int first_on=-1;
	char text[100];

	text[0]='\0';

	for(index=0; index < DTYPE_BITS; index++){
		if(num & (dtype)1<<63 /*DTYPE_LAST*/){
			strcat(text, "1");
			sp++;
			if(first_on==-1)
				first_on=index;
		}else if(sp){
			strcat(text, "0");
			sp++;
		}
		num<<=1;
		if(sp && ((first_on+sp) % 4)==0)
			strcat(text, " ");
		/*if(sp && (sp % 4)==0)
			strcat(text, " ");*/
	}
	fprintf(fout, "%s ", text);
}

/**
 * This method is a real time analyzer.
 */
void rt_analyzer(unsigned long *addr, unsigned long size, int ib){
	int tmu;
	int i;
	static unsigned long t_width=0, t_height=0;
	unsigned long e_magic1=0, e_magic2=0, magic1, magic2;
	unsigned long elements;
	unsigned char magic_tbl[4]={ 7, 4, 3, 0 };

	// Open a file named "rta" for writing
	if (fout == NULL) {
		fout = fopen("rta", "w");
		if (fout == NULL) {
			perror("rt_analyzer:");
			return;
		}
	}


	for(i = 0; i < size; i++) {
#if 1
		unsigned long prim_type;
		if(addr[i] == (RADEON_CP_PACKET3_3D_DRAW_INDX_2 | (0<<16)) ){
			i++;
			prim_type = addr[i] & 0xf;
			elements = addr[i] >> 16;

			e_magic1 = (8-((elements*2)%8))%8;
			e_magic2 = elements/2 + magic_tbl[(elements-1)%4];
		}
		//static o_idx_addr=0;
		if(addr[i] == (RADEON_CP_PACKET3_INDX_BUFFER | (2<<16)) ){
			unsigned long idx_addr, prim;
			//unsigned char magic1_tbl[4]={ 0, 6, 4, 2 };
			i++; magic1=(addr[i]>>16) & 0xff;
			prim=addr[i] & 0xffff;
			i++; idx_addr=addr[i];
			i++; magic2=addr[i];
			
			////dump_idx_buf_l(idx_addr, magic1*2/*elements*/);
			//dump_idx_buf_c(idx_addr, elements);
			//dump_idx_buf(idx_addr, elements);
			
			/*
			switch(prim_type){
				case R300_VAP_VF_CNTL__PRIM_TRIANGLES:
				case R300_VAP_VF_CNTL__PRIM_LINE_LOOP:
				case R300_VAP_VF_CNTL__PRIM_LINE_STRIP:
				case R300_VAP_VF_CNTL__PRIM_POINTS:
				case R300_VAP_VF_CNTL__PRIM_LINES:{
					*/
	
			fprintf(fout, "vcount %d, addr %x, magic1 %d, magic2 %d \n", elements, idx_addr, magic1, magic2);
			//fprintf(fout, "vcount %d, magic1 %d, magic2 %d \n", elements, magic1, magic2);
#if 0
			e_magic2=(elements+1)/2 + ((idx_addr&0x1c)/4);
					
			if(e_magic2 != magic2){
				fprintf(fout, "magic2 fails(%04lu vs %04lu)", magic2, e_magic2);
				//fprintf(fout, "e_magic2=%d ", e_magic2);
				//exit(0);
			}else
				fprintf(fout, "magic2 ok");
#endif					
			//fprintf(fout, "%s: ", prim_name(prim_type));

			//fprintf(fout, "\n");
			
			elements=0;
			prim_type=0;
		}
#endif

#if 1
		for(tmu=0; tmu < 16; tmu++){
			if(addr[i] == CP_PACKET0(R300_TX_SIZE_0+tmu*4,0)){
				i++;
				t_width=addr[i] & 0x7ff;
				t_height=(addr[i]>>11) & 0x7ff;
				t_width++;
				t_height++;
				continue;
			}
			if(addr[i] == CP_PACKET0(R300_TX_OFFSET_0+tmu*4,0)){
				i++;
				if(t_width && t_height /*&& (rand()%20)==0*/){
					//printf("dumping from %08x\n", addr[i]);
					check_texture(addr[i], t_width, t_height);
				}
			}
		}
#endif
	}
}
